
#include "includes.h"

//------------------------------------------------------------------------------
// Store error strings in flash to save RAM.
#define error(s) sd.errorHalt(&Serial, F(s))
//------------------------------------------------------------------------------

SdFs sd;                                                                        // Pointer to SD card
FsFile dir;                                                                     // Pointer to directory
FsFile file;                                                                    // Pointer to a file

uint8_t bytedata[DATA_BLOCK];                                                   // Array of data for loading the tape
LiquidCrystal_PCF8574 lcd(LCD_ADDRESS);                                         // Default address is set 

char aFileName[MAX_FILES][MAX_NAMELEN];                                         // Array to contain file names. It was set to have 255 files in a folder and 20 characters in file name
char aIsDir[MAX_FILES];                                                         // Array to store indication if it is a file or folder
char MyCurrentPath[MAX_PATH];                                                   // Current path where MAX_PATH is a maximum path len
uint32_t aFileCount;                                                            // How many files are in current folder
uint32_t aFptrNumOnScreen;                                                      // File number on screen
uint32_t aFptrNumOnDrive;                                                       // Base file number on drive, you have to add aFptrNumOnScreen to it to get file number
uint32_t aPrevStateBtn1 = 0;                                                    // Previous state of keyboard button 1
uint32_t aPrevStateBtn2 = 0;                                                    // Previous state of keyboard button 2
uint32_t aDebKey1 = 0;                                                          // Debouncer conuter for a Key1  
uint32_t aDebKey2 = 0;                                                          // Debouncer counter for a Key2
uint32_t cKey1Prev = 0;                                                         // Key1 previous state
uint32_t cKey2Prev = 0;                                                         // Key2 previous state
uint32_t tptr = 0;                                                              // Pointer to a currently loaded byte

void startTimer(int frequencyHz);                                               // Start timer function
void setTimerFrequency(int frequencyHz);                                        // Set timer frequency
void TC3_Handler();                                                             // Handler for timer 3

void setup() 
{
  pinMode(alxBTN_UP, INPUT);                                                    // Button up mode set
  pinMode(alxBTN_DN, INPUT);                                                    // Button down mode set
  pinMode(TAPE_OUT, OUTPUT);                                                    // Tape out
  pinMode(TURBO, OUTPUT);                                                       // Turbo loader pin  
  digitalWrite(TURBO, 1);                                                       // Disable turbo loading
  startTimer(0);                                                                // Starts timer
#if SERIAL_DEBUG == 1
  Serial.begin(115200);                                                         // Serial port initialization
#endif
  lcd.begin(LCD_CHARS, LCD_LINES);                                              // Initialize display
  lcd.setBacklight(LCD_BACKLT);                                                 // Set backlight as on    
  lcd.home();                                                                   // Reset cursor position
  lcd.clear();                                                                  // Clear it 
  lcd.createChar(ICN_FLDR, (uint8_t*)safldr);                                   // Folder icon
  lcd.createChar(ICN_FILE, (uint8_t*)safile);                                   // File icon
#if SERIAL_DEBUG == 1                                                           // if in debug mode
  while (!Serial)                                                               // Wait for serial port    
  {
    yield();                                                                    // Empty loop to wait for serial port to become available  
  }
  Serial.println("Type any character to start");                                // Once port is available
  while (!Serial.available())                                                   // Wait for any key to be pressed
  {
    yield();                                                                    // wait..  
  }
#endif  
}

void aLoadDir(char *aCurPath)                                                   // Function to load directory
{
  uint32_t uiDdot;                                                              // Flag to indicate that we are in subfolder
  uint32_t uiCntrFl;                                                            // Counter to anaalyze current path

  uiDdot = 0;                                                                   // Initialize variables
  uiCntrFl = 0;                                                                 // Initialize variables
  aFileCount = 0;                                                               // Initializes variables  

  memset((char*)aFileName, 0, sizeof(aFileName));                               // Clear the data where filenames are stored
  
  if (!sd.begin(SD_CS_PIN))                                                     // Open SD card  
  {
    sd.initErrorHalt(&Serial);                                                  // Yiled error if it did not open
  }

  if (!dir.open(aCurPath))                                                      // Trying to set path as was requested
  {
    error("dir.open failed");                                                   // Yiled an error if it was not possible    
  }

  while (aCurPath[uiCntrFl])                                                    // Searching for depath - if we are not in the root folder then we will need ".." characters to go up to be printed out
  {
    if (aCurPath[uiCntrFl++] == '/')                                            // Looking for separator
    {
      uiDdot++;                                                                 // And if found, then increment counter  
    }
  }

  if (uiDdot > 1)                                                               // If we are deeper than root
  {
    aFileCount = 1;                                                             // Then we need to insert ".." mark on top of the file list
    strcpy ((char *)aFileName, "..");                                           // Insert ".."
    aIsDir[0] = 1;                                                              // Set indication that first entry is a directory    
  }
    
  while (file.openNext(&dir, O_RDONLY))                                         // Trying to open next file to get its name
  {
    if (aFileCount < MAX_FILES)                                                 // Checking if we went thru less files than allocated memory for storage names
    {
      if (file.isDir())                                                         // If current file is a directory
      {
        aIsDir[aFileCount] = 1;                                                 // Then setting appropriate flag
      }
      else                                                                      // If it is not a directory
      {
        aIsDir[aFileCount] = 0;                                                 // Then also setting an appropriate flag  
      }
      file.getName(aFileName[aFileCount], MAX_NAMELEN);                         // Retrieve file name into the memory and limit size of the file name to MAX_NAMELEN
      file.close();                                                             // Close the file
      if (aFileName[aFileCount][0] > 32)                                        // If first character is a valid one (hidden filenames are starting from space)  
      {
        aFileCount++;                                                           // Then increment an array index
      }
    }
  }
  aSort((char*)aFileName, (char*)aIsDir, aFileCount);                           // Sort file list by name 
}

int strcasecmp(const char *s1, const char *s2)                                  // Function to compare strings caseless    
{                              
    const unsigned char *us1 = (const u_char *)s1,                              // Variables
                        *us2 = (const u_char *)s2;

    while (tolower(*us1) == tolower(*us2++))                                    // While loop using lower case                
        if (*us1++ == '\0')
            return (0);                                                         // If identical, return 0
    return (tolower(*us1) - tolower(*--us2));                                   // Otherwise return difference
}

void aSort(char *aArray, char *sDir, uint32_t aSizeOf)                          // Function to sort out folder contents
{
  uint32_t ui32Curpos;                                                          // Current sorted position
  uint32_t ui32Sorted;                                                          // Current position thru file list to be sorted
  char cTStr[MAX_NAMELEN];                                                      // Temporary storage for filename
  char chFOne[MAX_NAMELEN + 1];                                                 // Current smaple 1, increased size to hold space in the beginning for folder names. (Space in the beginning = folder)
  char chFTwo[MAX_NAMELEN + 1];                                                 // Current smaple 2, increased size to hold space in the beginning for folder names. (Space in the beginning = folder)
  char sIsDir;                                                                  // Indication if it is a folder    
  ui32Sorted = 0;                                                               // Make integers eq 0 at the beginning
  ui32Curpos = 0;                                                               // Make integers eq 0 at the beginning

  for(ui32Curpos = 0; ui32Curpos < aSizeOf; ui32Curpos++)                       // Outer loop of the sort algo to go thru entire array
  {
    for(ui32Sorted = ui32Curpos + 1; ui32Sorted < aSizeOf; ui32Sorted++)        // Inner loop of the sort algo to go thru unsorted part of the array
    {
      chFOne[0] = 0;                                                            // Make array "clear" by clearing the first byte of it
      if (sDir[ui32Curpos])                                                     // If current element is a folder 
      {
        strcat (chFOne, " ");                                                   // Then add a space in the beginning (folders first in the sorted list)
      }
      strcat (chFOne, &aArray[ui32Curpos * MAX_NAMELEN]);                       //  Copy the rest of the name

      chFTwo[0] = 0;                                                            // Same thing foir second argument
      if (sDir[ui32Sorted])                                                     // If it is a folder
      {
        strcat (chFTwo, " ");                                                   // Then add space in the beginning as folders shall come first in the sorted list
      }
      strcat (chFTwo, &aArray[ui32Sorted * MAX_NAMELEN]);                       // Copy the rest of the name
         
      if(strcasecmp(chFOne, chFTwo) > 0)                                        // If order is wrong, i.e. bbbb infront of aaaa then those have to be swapped
      {
        strcpy(cTStr, &aArray[ui32Curpos * MAX_NAMELEN]);                       // Copy value into temporary storage
        sIsDir = sDir[ui32Curpos];                                              // Copy "is it a folder" flag into temporary storage
                                                                                // Copy name from current location into the new location
        strcpy(&aArray[ui32Curpos * MAX_NAMELEN], &aArray[ui32Sorted * MAX_NAMELEN]);
        sDir[ui32Curpos] = sDir[ui32Sorted];                                    // Copy flag if it is a folder into the new location
        strcpy(&aArray[ui32Sorted * MAX_NAMELEN], cTStr);                       // Copy name from saved location
        sDir[ui32Sorted] = sIsDir;                                              // Copy folder flag from saved location
      }
     }
   }
#if SERIAL_DEBUG == 1                                                           // In case of serial debug is on
  for (ui32Sorted = 0; ui32Sorted < aSizeOf; ui32Sorted++)                      // print sorted array to a terminal
  {
    Serial.printf ("%s %d\r\n", &aArray[ui32Sorted*MAX_NAMELEN], (uint32_t)sDir[ui32Sorted]);
  }
#endif
}

void UpdateScreen(void)                                                         // Function to update current screen
{
  uint32_t CurLineCntr;                                                         // Current line counter
  uint32_t InLinePtr;                                                           // Current pointer within the line
  char cLineToP[LCD_CHARS+1];                                                   // Buffer to hold one line + EOL symbol
  CurLineCntr = 0;                                                              // Null the variables at the beginning
  InLinePtr = 0;                                                                // Null the variables at the beginning

  for (CurLineCntr = 0; CurLineCntr < LCD_LINES; CurLineCntr++)                 // Cycle thru all LCD lines
  {
    lcd.setCursor(0, CurLineCntr);                                              // Set cusror at the beginning of the line
    InLinePtr = 0;                                                              // Null the variable during each pass of the loop
    
    if (CurLineCntr == aFptrNumOnScreen)                                        // If we are to draw current selected file/folder
    {
      if (aIsDir[aFptrNumOnScreen + aFptrNumOnDrive] == 1)                      // Determine if this is going to be a file or folder
      {
        cLineToP[0] = '/';                                                      // Symbol to swhow if it is a folder
      }
      else
      {
        cLineToP[0] = '>';                                                      // Symbol to show if it is a file
      }
    }
    else 
    {
      cLineToP[0] = ' ';                                                        // Nothing, line is not where cursor is set
    }
    
    if (CurLineCntr + aFptrNumOnDrive >= MAX_FILES)                             // If we are at the end and line is empty
    {
      for (InLinePtr = 1; InLinePtr <= LCD_CHARS; InLinePtr++)                  // Make the line empty
      {
        cLineToP[InLinePtr] = ' ';                                              // Use spaces for that
      }
    }
    else 
    {
      for (InLinePtr = 1; InLinePtr < LCD_CHARS; InLinePtr++)                   // Otherwise use spaces at the end of the file name
      {
        if (aFileName[CurLineCntr + aFptrNumOnDrive][InLinePtr - 1] == 0)       // If 0 in the file name, it means past of the of the end of the name, so use space
        {
          while (InLinePtr < LCD_CHARS)                                         // Fill out the rest of the array with spaces  
          {
            cLineToP[InLinePtr++] = ' ';                                        // Use space
          }
        }
        else 
        {                                                                       // Otherwise use character from the file name
          cLineToP[InLinePtr] = aFileName[CurLineCntr + aFptrNumOnDrive][InLinePtr - 1];
        }
      }
    }
    cLineToP[LCD_CHARS] = 0;                                                    // End of the line shall be 0    
    if (cLineToP[0] == '>')                                                     // Check, if it is a file
    {
      lcd.print("\01");                                                         // Then print file icon
    }
    if (cLineToP[0] == '/')                                                     // Check if it is a folder
    {
      lcd.print("\02");                                                         // Then print a file icon  
    }
    if (cLineToP[0] == ' ')                                                     // Check if no icon shall be printed
    {
      lcd.printf(" ");                                                          // Print no icon
    }
    lcd.printf("%s", &cLineToP[1]);                                             // Print the name, either file or folder
  }
}

void ResetKeys(void)                                                            // Reset keyboard state machines
{
  aPrevStateBtn1 = 0;                                                           // Reset
  aPrevStateBtn2 = 0;                                                           // Reset
  aDebKey1 = 0;                                                                 // Reset
  aDebKey2 = 0;                                                                 // Reset
}

uint8_t KeyPressPoll(void)                                                      // Function to poll the keys. Idea is to raise an event only in case of button was released
{
  uint8_t aRet;                                                                 // Return value
  uint8_t CurrStateBtn1;                                                        // Current state button 1
  uint8_t CurrStateBtn2;                                                        // Current state button 2  

  aRet = 0;                                                                     // Make variables 0 at the beginning
  CurrStateBtn1 = 0;                                                            // Make variables 0 at the beginning
  CurrStateBtn2 = 0;                                                            // Make variables 0 at the beginning

  CurrStateBtn1 = digitalRead(alxBTN_UP);                                       // Poll the physical keys, button 'UP'
  CurrStateBtn2 = digitalRead(alxBTN_DN);                                       // Poll the physical keys, button 'DN'  
  
  if (CurrStateBtn1 != aPrevStateBtn1 && aDebKey1 == 0)                         // If state changed and debouncer 1 allows
  {
    aPrevStateBtn1 = CurrStateBtn1;                                             // Record the state of the button to the variable reflecting previous state
    aDebKey1 = 1;                                                               // Set debouncer flag 1 to disallow further key UP polling
  }

  if (aDebKey1 > 0)                                                             // If debouncer is enabled
  {
    aDebKey1++;                                                                 // Increment debouncer timer 
    if (aDebKey1 > KEYB_DELAY)                                                  // If timer expired
    {
      aDebKey1 = 0;                                                             // Clear the flag 
    }
  }
  
  if (aPrevStateBtn1 == 0)                                                      // If previous state is 0 then key is released 
  {              
    aRet &= KEY1_FLG;                                                           // Clear the bit
  }
  else                                                
  {
    aRet |= KEY1_FLP;                                                           // If previous state is 1, the bit needs to be set
  }

  if (CurrStateBtn2 != aPrevStateBtn2 && aDebKey2 == 0)                         // If state changed and debouncer 2 allows it
  {
    aPrevStateBtn2 = CurrStateBtn2;                                             // Record the state of the button to the variable reflecting previous state
    aDebKey2 = 1;                                                               // Set debouncer flag 2 to disallow further key DN polling
  }

  if (aDebKey2 > 0)                                                             // If debouncer is enabled  
  {
    aDebKey2++;                                                                 // Increment it
    if (aDebKey2 > KEYB_DELAY)                                                  // Check against the threshold
    {
      aDebKey2 = 0;                                                             // If it exceeds the threshold then set it to 0  (0 = disable debouncer)
    }
  }
  
  if (aPrevStateBtn2 == 0)                                                      // If previous state is 0 then key is released 
  {
    aRet &= KEY2_FLG;                                                           // Clear the flag
  }
  else 
  {
    aRet |= KEY2_FLP;                                                           // Otherwise set the flag  
  }
                                                                                // Sepcial case, two buttons are pressed simultaniously
  if (CurrStateBtn1 == 0 && CurrStateBtn2 == 0)                                 // If this condition is met 
  {
    aRet = KEY_SPCL;                                                            // Then set return flag appropriately 
  }
  
  return aRet;                                                                  // Return the result
}

uint8_t inkey(void)                                                             // Top level function to provide debounced key press values
{
  uint8_t aRet;                                                                 // Return value
  uint8_t cCurKeybState;                                                        // Current state 
  aRet = 0;                                                                     // Make variables 0 at the start
  cCurKeybState = 0;                                                            // Make variables 0 at the start

  cCurKeybState = KeyPressPoll();                                               // Poll keyboard
  if (cKey1Prev == 0 && (cCurKeybState & KEY1_FLP))                             // If previous state is released and current is pressed for key1
  {
    aRet |= KEY1_FLP;                                                           // Return value is key1
  }
  cKey1Prev = cCurKeybState & KEY1_FLP;                                         // Record the the state into the prevoius state
  
  if (cKey2Prev == 0 && (cCurKeybState & KEY2_FLP))                             // If previous state is released and current is pressed for key1
  {
    aRet |= KEY2_FLP;                                                           // Add key2 to return value
  }
  cKey2Prev = cCurKeybState & KEY2_FLP;                                         // And set previous state

  if (cCurKeybState == KEY_SPCL)                                                // If special combination of both buttons is pressed    
  {
    aRet = KEY1_FLP|KEY2_FLP;                                                   // Then return both keys pressed  and in this case there is no need to set any timers or do anything else
  }

  return aRet;                                                                  // Return the value
}
  
bool isFast(void)                                                               // Function to check filename for a .fst extention, and if ye, then this is a file for a fast loading mode
{
  bool bRet;                                                                    // Return bit, either false or true
  uint32_t DotFst;                                                              // Variable to search for .fst sequence  
  bRet = 0;                                                                     // Make variable null at the init
  DotFst = 0;                                                                   // Make variable null at the init

  for (DotFst = MAX_NAMELEN; DotFst; DotFst--)                                  // Loop from the end of the name, we need to find last 4 symbols if those are .fst
  {
    if (aFileName[aFptrNumOnScreen + aFptrNumOnDrive][DotFst - 1] == '.')       // If we found a dot
    {
                                                                                // the check 3 consequitive symbols to be "fst"
      if (aFileName[aFptrNumOnScreen + aFptrNumOnDrive][DotFst] == 'f' && aFileName[aFptrNumOnScreen + aFptrNumOnDrive][DotFst + 1] == 's' && aFileName[aFptrNumOnScreen + aFptrNumOnDrive][DotFst + 2] == 't'){
        bRet = 1;                                                               // And if found, then set return to be true
      }
    }
  }
  return bRet;                                                                  // Return the value
}

void loop(void)                                                                 // Main loop 
{
  uint8_t *curpos;                                                              // Pointer to a current position in the file being loaded
  uint8_t cKeyPress;                                                            // Keyboard polling variable
  uint32_t uiFileTSize;                                                         // Size of the file being loaded
  uint32_t uiFileLoadedPerTick;                                                 // File size expressed in 16384/2/18 ticks. 16384 = buffer size; 2 - two halfs of 1 ticks; 18 is spaces provided on the screen for slider
  uint8_t ui8TickCntr;                                                          // Screen ticks counter
  uint32_t uiPtrToCurP;                                                         // Pointer to current path position
  char strToLoad[1024];                                                         // File + path to be loaded

  uiFileTSize = 0;                                                              // Init variables at startup
  cKeyPress = 0;                                                                // Init variables at startup
  aFptrNumOnScreen = 0;                                                         // Init variables at startup    
  aFptrNumOnDrive = 0;                                                          // Init variables at startup
  uiPtrToCurP = 0;                                                              // Init variables at startup
  uiFileLoadedPerTick = 0;                                                      // Init variables at startup
  ui8TickCntr = 0;                                                              // Init variables at startup

  digitalWrite(TURBO, 1);                                                       // Set the turbo output to 1, meaning normal loading speed
  strcpy (MyCurrentPath, "/");                                                  // Set the root path for a start condition '/' 
  aLoadDir(MyCurrentPath);                                                      // Load directory at the current path    
  
  while (inkey());                                                              // Clear keypad buffer  

  UpdateScreen();                                                               // Update screen with loaded directory
  while (true)                                                                  // Inifinite loop, there is no exit from main
  {
    cKeyPress = inkey();                                                        // Poll for key press
    if (cKeyPress != 0)                                                         // If key has been pressed
    {
      if (cKeyPress == KEYB_UP)                                                 // If it is Key UP
      {             
        if (aFptrNumOnScreen == 0)                                              // If we are at the top of display already
        {
          if (aFptrNumOnDrive != 0)                                             // If we are not at the top of the directory
          {
            aFptrNumOnDrive--;                                                  // Decrese current selection for on screen
            UpdateScreen();                                                     // Update screen
          }
        }
        else 
        {
          aFptrNumOnScreen--;                                                   // Otherwise if awe are not of the display then decrease selected number on screen
          UpdateScreen();                                                       // Update screen    
        }
      }
                                                                                // If Key DN has been pressed and there is a way to go down
      if (cKeyPress == KEYB_DN && aFptrNumOnScreen + aFptrNumOnDrive < aFileCount - 1) 
      {
        if (aFptrNumOnScreen >= LCD_LINES - 1)                                  // If number selected is at the bottom
        {
          if (aFptrNumOnDrive < aFileCount - LCD_LINES)                         // If there are more files
          {
            aFptrNumOnDrive++;                                                  // Then move the selector
            UpdateScreen();                                                     // Update the screen
          }
        }
        else                                                                    // If number selected is not at the bottom
        {
          if (aFptrNumOnScreen + aFptrNumOnDrive <= aFileCount)                 // If there are more files, so operation is possible
          {
            aFptrNumOnScreen++;                                                 // Then move the selector
            UpdateScreen();                                                     // Update screen
          }
        }
      }

      if (cKeyPress == KEYB_EN)                                                 // If two button pressed together
      {
        if (aIsDir[aFptrNumOnScreen + aFptrNumOnDrive] == 0)                    // If selection is not a folder
        {
          lcd.clear();                                                          // Clear screen
          lcd.printf("File loading:");                                          // Print the banner
          lcd.setCursor(0, 1);                                                  // Print the banner
          lcd.printf("%s", aFileName[aFptrNumOnScreen + aFptrNumOnDrive]);      // Print the banner
          while (inkey());                                                      // Wait for all keys to be released  
          strcpy (strToLoad, MyCurrentPath);                                    // Build a complete path with a filename out of just file path
          strcat (strToLoad, aFileName[aFptrNumOnScreen + aFptrNumOnDrive]);    // Build a complete path with a filename out of just file path  
          if (file.open(strToLoad))                                             // If it is possible to open a file     
          {
            uiFileTSize = (file.size() / SIZE_HALF) / LCD_TICKS;                // Count the tick size based on file size     
            ui8TickCntr = 0;                                                    // Reset tick counter
            
            if (isFast()){                                                      // Check wether it is fast file or not
              digitalWrite(TURBO, 0);                                           // If fast, then set h/w flag to indicate to ZX FPGA that fast mode has been selected
              lcd.setCursor(0, 2);                                              // Set cursor position
              lcd.printf("Mode: fast");                                         // And print "Mode: fast"
            }
            else                                                                // Otherwise, if it is not a fast mode 
            {
              lcd.setCursor(0, 2);                                              // Set cursor position and print so
              lcd.printf("Mode: normal");                                       // "Mode: normal"
            }
            lcd.setCursor(0, 3);                                                // Display slider
            lcd.printf ("[                  ]");                                // Empty at the beginning
            lcd.setCursor(1, 3);                                                // Position the cursor  
            while (file.available())                                            // Loading
            {
              if (tptr == SIZE_HALF)                                            // If we reach with loading the upper half of the buffer
              {
                curpos = bytedata;                                              // Set current position pointer to point to memeory location where it will be sent to "tape in" from, currently at 0 as we need upper portion to be loaded first     
                file.read(curpos, SIZE_HALF);                                   // Read the portion of the file  
                uiFileLoadedPerTick++;                                          // Increment tick counter
                if (uiFileLoadedPerTick >= uiFileTSize)                         // Check if new tick is required
                {
                  uiFileLoadedPerTick = 0;                                      // If so, reset the counter
                  if (ui8TickCntr < LCD_TICKS)                                  // Precaution - if space left then print the tick
                  {
                    lcd.printf("-");                                            // Print the tick
                  }
                  ui8TickCntr ++;                                               // Increment tick counter
                }
                while (tptr < UPPER_THRESH);                                    // Wait till it moves
              }
              if (tptr == 0)                                                    // If we are at the lower half of the memory to be filled in by the data to be loaded to "tape in"
              {
                curpos = SIZE_HALF + bytedata;                                  // Load the upper half of the memory (while lower is being laoded to the mic)
                file.read(curpos, SIZE_HALF);                                   // Read the portion of the file
                uiFileLoadedPerTick++;
                if (uiFileLoadedPerTick >= uiFileTSize)                         // Check if new tick is required
                {
                  uiFileLoadedPerTick = 0;                                      // If so, reset the counter
                  if (ui8TickCntr < LCD_TICKS)                                  // Precaution - if space left then print the tick
                  {
                    lcd.printf("-");                                            // Print the tick
                  }
                  ui8TickCntr ++;                                               // Increment tick counter
                }
                while (tptr < LOWER_THRESH);                                    // Wait till it moves
              }
            }
            file.close();                                                       // Close the file as it was loaded
            delay(100);                                                         // Small delay to ensure that we finish downloading process
            digitalWrite(TURBO, 1);                                             // Clear the turbo flag
            while(inkey() != 0);                                                // Wait for key to be released
            lcd.clear();                                                        // Clear the LCD  
            lcd.printf("Loaded:");                                              // Print "loaded"
            lcd.setCursor(0, 1);                                                // Set the position
            lcd.printf("%s", aFileName[aFptrNumOnScreen + aFptrNumOnDrive]);    // Prints the filename
            lcd.setCursor(0, 3);                                                // Set the position
            lcd.printf ("[------------------]");                                // Print full line as precaution
            delay(1000);                                                        // Short pause
            ResetKeys();                                                        // Reset keyboard
            while(inkey() == 0);                                                // Wait for key to be pressed
          }
          else                                                                  // Otherwise "R Tape loading error" :) condition
          {
            lcd.setCursor(0, 2);                                                // Set the lcd position
            lcd.printf("R Tape Loading Error");                                 // Print the message
          }
        }
        else                                                                    // Otherwise folder is selcted so we need to descend to it
        {
                                                                                // if ".." folder is selcted  
          if (aFileName[aFptrNumOnScreen + aFptrNumOnDrive][0] == '.' && aFileName[aFptrNumOnScreen + aFptrNumOnDrive][1] == '.')
          {
                                                                                // Strip the last folder from the name
            for (uiPtrToCurP = 0; uiPtrToCurP < strlen(MyCurrentPath); uiPtrToCurP++)
            {
              if (MyCurrentPath[strlen(MyCurrentPath) - uiPtrToCurP - 2] == '/')// If folder separator is found i.e. "/"
              {
                MyCurrentPath[strlen(MyCurrentPath) - uiPtrToCurP - 2] = 0;     // Zero it, so string will become "one folder less"
                uiPtrToCurP = strlen(MyCurrentPath) + 1;                        // End the loop prematurely            
              }
            }
          }
          else                                                                  // If regular folder is selected
          {
                                                                                // Add it to the path
            strcat (MyCurrentPath, aFileName[aFptrNumOnScreen + aFptrNumOnDrive]);
          }
          strcat (MyCurrentPath, "/");                                          // Add "/" at the end of the path
          uiFileTSize = 0;                                                       // Reset all variables
          cKeyPress = 0;                                                        // Reset all variables
          aFptrNumOnScreen = 0;                                                 // Reset all variables
          aFptrNumOnDrive = 0;                                                  // Reset all variables
          aLoadDir(MyCurrentPath);                                              // Load new directory
          lcd.clear();                                                          // Clear screen    
          UpdateScreen();                                                       // Update screen
          ResetKeys();                                                          // Reset keyboard
          delay(1000);                                                          // Small delay
          while (inkey() != 0);                                                 // Wait for keys to be released
        }
      }
    }
  }
}

void setTimerFrequency(int frequencyHz) 
{
  int compareValue = 8;                                                         // Originnaly 16, changed to 8. (CPU_HZ / (TIMER_PRESCALER_DIV * frequencyHz)) - 1;
  TcCount16* TC = (TcCount16*) TC3;                                             // Timer 3 counter
  TC->COUNT.reg = map(TC->COUNT.reg, 0, TC->CC[0].reg, 0, compareValue);        // Load hw register
  TC->CC[0].reg = compareValue;                                                 // Load hw register  
  while (TC->STATUS.bit.SYNCBUSY == 1);                                         // Wait till bit is cleared
}

void startTimer(int frequencyHz) 
{
                                                                                // Set the REG_GCLK_CLKCTRL with appropriate bits 
  REG_GCLK_CLKCTRL = (uint16_t) (GCLK_CLKCTRL_CLKEN | GCLK_CLKCTRL_GEN_GCLK0 | GCLK_CLKCTRL_ID_TCC2_TC3) ;
  while ( GCLK->STATUS.bit.SYNCBUSY == 1 );                                     // wait for sync

  TcCount16* TC = (TcCount16*) TC3;                                             // Timer 3 counter  

  TC->CTRLA.reg &= ~TC_CTRLA_ENABLE;                                            // Set bits
  while (TC->STATUS.bit.SYNCBUSY == 1);                                         // wait for sync

  TC->CTRLA.reg |= TC_CTRLA_MODE_COUNT16;                                       // Use the 16-bit timer
  while (TC->STATUS.bit.SYNCBUSY == 1);                                         // wait for sync

  TC->CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;                                       // Use match mode so that the timer counter resets when the count matches the compare register
  while (TC->STATUS.bit.SYNCBUSY == 1);                                         // wait for sync

  TC->CTRLA.reg |= TC_CTRLA_PRESCALER_DIV64;                                    // Set prescaler to 64
  while (TC->STATUS.bit.SYNCBUSY == 1);                                         // wait for sync

  setTimerFrequency(frequencyHz);                                               // Set timer frequency

  TC->INTENSET.reg = 0;                                                         // Enable the compare interrupt
  TC->INTENSET.bit.MC0 = 1;                                                     // Enable the compare interrupt  

  NVIC_EnableIRQ(TC3_IRQn);                                                     // Enable IRQ

  TC->CTRLA.reg |= TC_CTRLA_ENABLE;                                             // Enable it
  while (TC->STATUS.bit.SYNCBUSY == 1);                                         // wait for sync
}

void TC3_Handler(void)                                                          // Actual timer handler 
{
  TcCount16* TC = (TcCount16*) TC3;                                             // poniter to timer
  if (TC->INTFLAG.bit.MC0 == 1)                                                 // If we are in timer interrupt mode 
  {
    TC->INTFLAG.bit.MC0 = 1;                                                    // Then clear interrupt flag
    if (bytedata[tptr] == 0xff)                                                 // If current data is 0xff
    {
      digitalWrite(TAPE_OUT, 0);                                                // then clear the bit
    }
    else
    {
      digitalWrite(TAPE_OUT, 1);                                                // otherwise set the bit
    }
    tptr++;                                                                     // increment pointer
    if (tptr >= 16384)                                                          // If pointer is more than 16K  
    {
      tptr = 0;                                                                 // reset it
    }
  }
}
